// ----------------------------------------------------------------
// LCD Display Funktionen, die mit dem 2 x 16 LCD Display arbeiten
// ----------------------------------------------------------------

#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include "lcd.h"

#define F_CPU 20000000UL 
#include <util/delay.h>


// constants/macros
#define DDR(x) (*(&x - 1))      /* address of data direction register of port x */



#if defined(__AVR_ATmega64__) || defined(__AVR_ATmega128__)
   /* on ATmega64/128 PINF is on port 0x00 and not 0x60 */
   #define PIN(x) ( &PORTF==&(x) ? _SFR_IO8(0x00) : (*(&x - 2)) )
#else
       #define PIN(x) (*(&x - 2))    /* address of input register of port x          */
#endif
//


#if LCD_IO_MODE
#define lcd_e_delay()   __asm__ __volatile__( "rjmp 1f\n 1:" );
#define lcd_e_high()    PORTC  |=  _BV(PORTC2);
#define lcd_e_low()     PORTC  &= ~_BV(PORTC2);
#define lcd_e_toggle()  toggle_e()
#define lcd_rw_high()   PORTC |=  _BV(PORTC1)
#define lcd_rw_low()    PORTC &= ~_BV(PORTC1)
#define lcd_rs_high()   PORTC |=  _BV(PC0)
#define lcd_rs_low()    PORTC &= ~_BV(PC0)
#endif
//

#if LCD_IO_MODE
#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_1LINE
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_4BIT_2LINES
#endif
#else
#if LCD_LINES==1
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_8BIT_1LINE
#else
#define LCD_FUNCTION_DEFAULT    LCD_FUNCTION_8BIT_2LINES
#endif
#endif
//


//function prototypes
#if LCD_IO_MODE
static void toggle_e(void);
#endif
//



//-------------------------------------------------------------------------------------------
// Lokale Funktionen
//-------------------------------------------------------------------------------------------




/*************************************************************************
delay loop fr kleine genaue delays: 16-bit counter, 4 cycles/loop
*************************************************************************/
static inline void _delayFourCycles(unsigned int __count)
{
   if ( __count == 0 )    
       __asm__ __volatile__( "rjmp 1f\n 1:" );    // 2 cycles
   else
       __asm__ __volatile__ (
           "1: sbiw %0,1" "\n\t"                  
           "brne 1b"                              // 4 cycles/loop
           : "=w" (__count)
           : "0" (__count)
          );
}
//



//**********************************************************************************
// delay fr ein minimum von <us> microseconds
// Die Anzahl der Loops wird beim compile-time aus der MCU clock frequency berechnet
//***********************************************************************************
#define delay(us)  _delayFourCycles( ( ( 1*(XTAL/4000) )*us)/1000 )


#if LCD_IO_MODE
// toggled das Enable Pin, um Schreiben einzuleiten
static void toggle_e(void)
{
   lcd_e_high();
   lcd_e_delay();
   lcd_e_low();
}
#endif
//



/*************************************************************************
Function, um ein Byte auf dem LCD Controller zu schreiben
Input:    data   byte to write to LCD
         rs     1: write data    
				0: write instruction
Returns:  none
*************************************************************************/
#if LCD_IO_MODE
void lcd_write(uint8_t data,uint8_t rs)
{
   unsigned char dataBits ;


   if (rs) {   /* write data        (RS=1, RW=0) */
      lcd_rs_high();
   } else {    /* write instruction (RS=0, RW=0) */
      lcd_rs_low();
   }
   lcd_rw_low();

   if ( (PORTC4 == 0) && (PORTC5 == 1) && (PORTC6 == 2) && (PORTC7 == 3) )
   {
       /* configure data pins as output */
       DDR(PORTC) |= 0x0F;

       /* output high nibble first */
       dataBits = PORTC & 0xF0;
       PORTC = dataBits |((data>>4)&0x0F);
       lcd_e_toggle();

       /* output low nibble */
       PORTC = dataBits | (data&0x0F);
       lcd_e_toggle();

       /* all data pins high (inactive) */
       PORTC = dataBits | 0x0F;
   }
   else
   {
       /* configure data pins as output */
       DDR(PORTC) |= _BV(PORTC4);
       DDR(PORTC) |= _BV(PORTC5);
       DDR(PORTC) |= _BV(PORTC6);
       DDR(PORTC) |= _BV(PORTC7);
       
       /* output high nibble first */
       PORTC &= ~_BV(PORTC7);
       PORTC &= ~_BV(PORTC6);
       PORTC &= ~_BV(PORTC5);
       PORTC &= ~_BV(PORTC4);
       if(data & 0x80) PORTC |= _BV(PORTC7);
       if(data & 0x40) PORTC |= _BV(PORTC6);
       if(data & 0x20) PORTC |= _BV(PORTC5);
       if(data & 0x10) PORTC |= _BV(PORTC4);  
       lcd_e_toggle();
       
       /* output low nibble */
       PORTC &= ~_BV(PORTC7);
       PORTC &= ~_BV(PORTC6);
       PORTC &= ~_BV(PORTC5);
       PORTC &= ~_BV(PORTC4);
       if(data & 0x08) PORTC |= _BV(PORTC7);
       if(data & 0x04) PORTC |= _BV(PORTC6);
       if(data & 0x02) PORTC |= _BV(PORTC5);
       if(data & 0x01) PORTC |= _BV(PORTC4);
       lcd_e_toggle();        
       
       /* all data pins high (inactive) */
       PORTC |= _BV(PORTC4);
       PORTC |= _BV(PORTC5);
       PORTC |= _BV(PORTC6);
       PORTC |= _BV(PORTC7);
   }
}
#else
#define lcd_write(d,rs) if (rs) *(volatile uint8_t*)(LCD_IO_DATA) = d; else *(volatile uint8_t*)(LCD_IO_FUNCTION) = d;
/* rs==0 -> write instruction to LCD_IO_FUNCTION */
/* rs==1 -> write data to LCD_IO_DATA */
#endif
//



/*************************************************************************
Low-level function to read byte from LCD controller
Input:    rs     1: read data    
                0: read busy flag / address counter
Returns:  byte read from LCD controller
*************************************************************************/
#if LCD_IO_MODE
static uint8_t lcd_read(uint8_t rs)
{
   uint8_t data;
   
   
   if (rs)
       lcd_rs_high();                       /* RS=1: read data      */
   else
       lcd_rs_low();                        /* RS=0: read busy flag */
   lcd_rw_high();                           /* RW=1  read mode      */
   
   if ( ( &PORTC == &PORTC) && ( &PORTC == &PORTC ) && ( &PORTC == &PORTC )
     && ( PORTC4 == 0 )&& (PORTC5 == 1) && (PORTC6 == 2) && (PORTC7 == 3) )
   {
       DDR(PORTC) &= 0xF0;         /* configure data pins as input */
       
       lcd_e_high();
       lcd_e_delay();        
       data = PIN(PORTC) << 4;     /* read high nibble first */
       lcd_e_low();
       
       lcd_e_delay();                       /* Enable 500ns low       */
       
       lcd_e_high();
       lcd_e_delay();
       data |= PIN(PORTC)&0x0F;    /* read low nibble        */
       lcd_e_low();
   }
   else
   {
       /* configure data pins as input */
       DDR(PORTC) &= ~_BV(PORTC4);
       DDR(PORTC) &= ~_BV(PORTC5);
       DDR(PORTC) &= ~_BV(PORTC6);
       DDR(PORTC) &= ~_BV(PORTC7);
               
       /* read high nibble first */
       lcd_e_high();
       lcd_e_delay();        
       data = 0;
       if ( PIN(PORTC) & _BV(PORTC4) ) data |= 0x10;
       if ( PIN(PORTC) & _BV(PORTC5) ) data |= 0x20;
       if ( PIN(PORTC) & _BV(PORTC6) ) data |= 0x40;
       if ( PIN(PORTC) & _BV(PORTC7) ) data |= 0x80;
       lcd_e_low();

       lcd_e_delay();                       /* Enable 500ns low       */
   
       /* read low nibble */    
       lcd_e_high();
       lcd_e_delay();
       if ( PIN(PORTC) & _BV(PORTC4) ) data |= 0x01;
       if ( PIN(PORTC) & _BV(PORTC5) ) data |= 0x02;
       if ( PIN(PORTC) & _BV(PORTC6) ) data |= 0x04;
       if ( PIN(PORTC) & _BV(PORTC7) ) data |= 0x08;        
       lcd_e_low();
   }
   return data;
}
#else
#define lcd_read(rs) (rs) ? *(volatile uint8_t*)(LCD_IO_DATA+LCD_IO_READ) : *(volatile uint8_t*)(LCD_IO_FUNCTION+LCD_IO_READ)
/* rs==0 -> read instruction from LCD_IO_FUNCTION */
/* rs==1 -> read data from LCD_IO_DATA */
#endif
//



static uint8_t lcd_waitbusy(void)
/*************************************************************************
Loopt whrend LCD busy ist, address counter kommt auf return zurck
*************************************************************************/
{
   register uint8_t c;
   
   /* wait until busy flag is cleared */
   while ( (c=lcd_read(0)) & (1<<LCD_BUSY)) {}
   
   /* the address counter is updated 4us after the busy flag is cleared */
   delay(2);

   /* now read the address counter */
   return (lcd_read(0));  // return address counter
   
}
// ------------------------- end of lcd_waitbusy ------------------------



static inline void lcd_newline(uint8_t pos)
/*************************************************************************
Cursor springt auf die nchste Zeile
*************************************************************************/
{
   register uint8_t addressCounter;

#if LCD_LINES==1
   addressCounter = 0;
#endif
#if LCD_LINES==2
   if ( pos < (LCD_START_LINE2) )
       addressCounter = LCD_START_LINE2;
   else
       addressCounter = LCD_START_LINE1;
#endif
#if LCD_LINES==4
#if KS0073_4LINES_MODE
   if ( pos < LCD_START_LINE2 )
       addressCounter = LCD_START_LINE2;
   else if ( (pos >= LCD_START_LINE2) && (pos < LCD_START_LINE3) )
       addressCounter = LCD_START_LINE3;
   else if ( (pos >= LCD_START_LINE3) && (pos < LCD_START_LINE4) )
       addressCounter = LCD_START_LINE4;
   else
       addressCounter = LCD_START_LINE1;
#else
   if ( pos < LCD_START_LINE3 )
       addressCounter = LCD_START_LINE2;
   else if ( (pos >= LCD_START_LINE2) && (pos < LCD_START_LINE4) )
       addressCounter = LCD_START_LINE3;
   else if ( (pos >= LCD_START_LINE3) && (pos < LCD_START_LINE2) )
       addressCounter = LCD_START_LINE4;
   else
       addressCounter = LCD_START_LINE1;
#endif
#endif
   lcd_command((1<<LCD_DDRAM)+addressCounter);
}
//------------------------------------ end of lcd_newline ---------------------





// ----------------------------------------------------------------------------------
// ffentliche Funktionen
// ----------------------------------------------------------------------------------


void lcd_command(uint8_t cmd)
// *************************************************************************
// Send 		LCD Controller Kommando
// Input:   	Instruction zum Senden an den LCD-Controller
// Return: 		void
// ------------------------------------------------------------------------
{
   lcd_waitbusy();
   lcd_write(cmd,0);
}
//*************************************************************************



void lcd_data(uint8_t data)
//*************************************************************************
// Sende 		Datenbyte an den LCD controller
// Input:   	data to send to LCD controller, see HD44780 data sheet
// Return: 		void
// ------------------------------------------------------------------------
{
   lcd_waitbusy();
   lcd_write(data,1);
}
//*************************************************************************



void lcd_gotoxy(uint8_t x, uint8_t y)
//*************************************************************************
// Setze 		Cursor auf eine bestimmte Position
// Input:    	x  horizontal position  (0: ganz links)
//           	y  vertical position    (0: erste Zeile)
// Returns:  	void
// ------------------------------------------------------------------------
{
#if LCD_LINES==1
   lcd_command((1<<LCD_DDRAM)+LCD_START_LINE1+x);
#endif
#if LCD_LINES==2
   if ( y==0 )
       lcd_command((1<<LCD_DDRAM)+LCD_START_LINE1+x);
   else
       lcd_command((1<<LCD_DDRAM)+LCD_START_LINE2+x);
#endif
}
//*************************************************************************



//******************************************************************************
//                      Liefert die Position des Cursurs
// -----------------------------------------------------------------------------
int lcd_getxy(void)
{
   return lcd_waitbusy();
}
//******************************************************************************



//******************************************************************************
//                     Lscht das LCD-Display und Setzt Cursor auf die Home-Position
//						(funktioniert aber noch nicht)
void lcd_clrscr(void)
{
   lcd_command(0x01);			
}
//******************************************************************************



//*************************************************************************
//                      Setzt Cursor auf die Home-Position
void lcd_home(void)					// #define LCD_HOME              1      /* DB1: return to home position        */
{
   lcd_command(1<<LCD_HOME);
}
//***********************************************************************


void lcd_putc(char c)
/*************************************************************************
Character 	auf der aktuellen Cursorposition anzeigen
Input:    	char                                     
Return:		void  
*************************************************************************/
{
   uint8_t pos;

   pos = lcd_waitbusy();   // read busy-flag and address counter
   if (c=='\n')
   {
       lcd_newline(pos);
   }
   else
   {
#if LCD_WRAP_LINES==1
#if LCD_LINES==1
       if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE1,0);
       }
#elif LCD_LINES==2
       if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE2,0);    
       }else if ( pos == LCD_START_LINE2+LCD_DISP_LENGTH ){
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE1,0);
       }
#elif LCD_LINES==4
       if ( pos == LCD_START_LINE1+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE2,0);    
       }else if ( pos == LCD_START_LINE2+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE3,0);
       }else if ( pos == LCD_START_LINE3+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE4,0);
       }else if ( pos == LCD_START_LINE4+LCD_DISP_LENGTH ) {
           lcd_write((1<<LCD_DDRAM)+LCD_START_LINE1,0);
       }
#endif
       lcd_waitbusy();
#endif
       lcd_write(c, 1);
   }

}
// ---------------------- end of lcd_putc -------------------------------------




void lcd_puts(const char *s)
/*************************************************************************
Display string without auto linefeed
Input:    string to be displayed
Returns:  none
*************************************************************************/
{
   register char c;

   while ( (c = *s++) ) {
       lcd_putc(c);
   }

}
// ---------------------- end of lcd_puts -------------------------------------



void lcd_puts_p(const char *progmem_s)
/*************************************************************************
Display string from program memory without auto linefeed
Input:     string from program memory be be displayed                                        
Returns:   none
*************************************************************************/
/* print string from program memory on lcd (no auto linefeed) */
{
   register char c;

   while ( (c = pgm_read_byte(progmem_s++)) ) {
       lcd_putc(c);
   }

}
// ---------------------- end of lcd_puts_p -------------------------------




void lcd_init(uint8_t dispAttr)
/*************************************************************************
Initialisieren des LCD Displays
Input:    dispAttr 	LCD_DISP_OFF            display off
					LCD_DISP_ON             display on, cursor off
					LCD_DISP_ON_CURSOR      display on, cursor on
					LCD_DISP_CURSOR_BLINK   display on, cursor on flashing
Returns:  none
*************************************************************************/
{
#if LCD_IO_MODE
   /*
    *  Initialize LCD to 4 bit I/O mode
    */
   
   if ( (PORTC4 == 0 ) && (PORTC5 == 1) && (PORTC6 == 2) && (PORTC7 == 3)
     && (PC0 == 4 ) && (PORTC1 == 5) && (PORTC2 == 6 ) )
   {
       /* configure all port bits as output (all LCD lines on same port) */
       DDR(PORTC) |= 0x7F;
   }
   else if ( (PORTC4 == 0 ) && (PORTC5 == 1) && (PORTC6 == 2) && (PORTC7 == 3) )
   {
       /* configure all port bits as output (all LCD data lines on same port, but control lines on different ports) */
       DDR(PORTC) |= 0x0F;
       DDR(PORTC) |= _BV(PC0);
       DDR(PORTC) |= _BV(PORTC1);
       DDR(PORTC) |= _BV(PORTC2);
   }
   else
   {
       /* configure all port bits as output (LCD data and control lines on different ports */
       DDR(PORTC) |= _BV(PC0);
       DDR(PORTC) |= _BV(PORTC1);
       DDR(PORTC) |= _BV(PORTC2);
       DDR(PORTC) |= _BV(PORTC4);
       DDR(PORTC) |= _BV(PORTC5);
       DDR(PORTC) |= _BV(PORTC6);
       DDR(PORTC) |= _BV(PORTC7);
   }
   _delay_ms(16);        /* wait 16ms or more after power-on       */
   
   /* initial write to lcd is 8bit */
   PORTC |= _BV(PORTC5);  // _BV(LCD_FUNCTION)>>4;
   PORTC |= _BV(PORTC4);  // _BV(LCD_FUNCTION_8BIT)>>4;
   lcd_e_toggle();
   delay(4992);         /* delay, busy flag can't be checked here */
 
   /* repeat last command */
   lcd_e_toggle();      
   delay(64);           /* delay, busy flag can't be checked here */
   
   /* repeat last command a third time */
   lcd_e_toggle();      
   delay(64);           /* delay, busy flag can't be checked here */

   /* now configure for 4bit mode */
   PORTC &= ~_BV(PORTC4);   // LCD_FUNCTION_4BIT_1LINE>>4
   lcd_e_toggle();
   delay(64);           /* some displays need this additional delay */
   
   /* from now the LCD only accepts 4 bit I/O, we can use lcd_command() */    
#else
   /*
    * Initialize LCD to 8 bit memory mapped mode
    */
   
   /* enable external SRAM (memory mapped lcd) and one wait state */        
   MCUCR = _BV(SRE) | _BV(SRW);

   /* reset LCD */
   delay(16000);                           /* wait 16ms after power-on     */
   lcd_write(LCD_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                  
   delay(4992);                            /* wait 5ms                     */
   lcd_write(LCD_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                
   delay(64);                              /* wait 64us                    */
   lcd_write(LCD_FUNCTION_8BIT_1LINE,0);   /* function set: 8bit interface */                
   delay(64);                              /* wait 64us                    */
#endif

#if KS0073_4LINES_MODE
   /* Display with KS0073 controller requires special commands for enabling 4 line mode */
       lcd_command(KS0073_EXTENDED_FUNCTION_REGISTER_ON);
       lcd_command(KS0073_4LINES_MODE);
       lcd_command(KS0073_EXTENDED_FUNCTION_REGISTER_OFF);
#else
   lcd_command(LCD_FUNCTION_DEFAULT);      /* function set: display lines  */
#endif
   lcd_command(LCD_DISP_OFF);              /* display off                  */
   lcd_clrscr();                           /* display clear                */
   lcd_command(LCD_MODE_DEFAULT);          /* set entry mode               */
   lcd_command(dispAttr);                  /* display/cursor control       */

}
//***************************************************************************************************
